import time
import uuid
import exceptions
import importlib
import json
import yaml
import socket
import datetime

from twisted.application import service
from twisted.internet import task, reactor, protocol, defer
from twisted.python import log

from txredis.client import RedisClient

from rhumba.queue import RhumbaQueue
from rhumba.http_api import APIService

class RhumbaService(service.Service):
    """ Rhumba service
    Runs timers, configures sources and and manages the queue
    """
    def __init__(self, config):
        try:
            self.config = yaml.load(config)
        except:
            self.config = {}

        self.uuid = uuid.uuid1().get_hex()

        self.backend = self.config.get('backend', 'rhumba.backends.redis')

        self.api_service = self.config.get('api_enabled', True)

        self.hostname = self.config.get('hostname')
        if not self.hostname:
            hs = socket.gethostname()
            if '.' in hs:
                self.hostname = hs
            else:
                # Try and resolve our FQDN
                try:
                    self.hostname = socket.gethostbyaddr(hs)[0]
                except:
                    self.hostname = hs

        self.redis_host = self.config.get('redis_host', 'localhost')
        self.redis_port = int(self.config.get('redis_port', 6379))

        self.expire = 3600
        
        self.queues = {}

        self.crons = {}

        self.t = task.LoopingCall(self.heartbeat)

    @defer.inlineCallbacks
    def checkCrons(self, now):
        for queue, crons in self.crons.items():
            runner = yield self.checkCronRunners(queue)

            if not runner:
                yield self.registerCronRunner(queue)
                runner = self.uuid

            if runner == self.uuid:
                for cron in crons:
                    lastRun = yield self.lastRun(queue, cron.name)

                    if cron.checkCron(lastRun, now):
                        plug = self.queues[queue].plugin
                        yield self.setLastRun(queue, cron.name, now)

                        # Queue this job
                        yield self.client.queue(
                            queue, cron.name.split('call_', 1)[-1])
   
    @defer.inlineCallbacks
    def heartbeat(self):
        yield self.checkCrons(datetime.datetime.now())

        yield self.client.set("rhumba.server.%s.uuid" % self.hostname,
            self.uuid, expire=self.expire)

        yield self.client.set("rhumba.server.%s.heartbeat" % self.hostname,
            time.time(), expire=self.expire)

        yield self.client.set("rhumba.server.%s.queues" % self.hostname,
            json.dumps(self.queues.keys()), expire=self.expire)

    def setStatus(self, status):
        return self.client.set("rhumba.server.%s.status" % self.hostname,
            status, expire=self.expire)

    def startBeat(self):
        self.td = self.t.start(1.0)

    def lastRun(self, queue, fn):
        return self.client.get("rhumba.crons.%s.%s" % (queue, fn))

    def setLastRun(self, queue, fn, now):
        now = time.mktime(now.timetuple())
        return self.client.set("rhumba.crons.%s.%s" % (queue, fn), now)

    def registerCronRunner(self, queue):
        return self.client.set("rhumba.crons.%s" % queue, self.uuid, expire=60)

    def deregisterCronRunner(self, queue):
        return self.client.delete("rhumba.crons.%s" % queue)

    def checkCronRunners(self, queue):
        return self.client.get("rhumba.crons.%s" % queue)

    def setupQueues(self):
        queues = self.config.get('queues', [])
        for queue in queues:
            q = RhumbaQueue(queue, self)
            qname = queue['name']
            self.queues[qname] = q

            if q.plugin:
                # Find all methods decorated with @cron
                crons = [
                    getattr(q.plugin, i).cronable for i in dir(q.plugin)
                    if hasattr(getattr(q.plugin, i), 'cronable')
                ]

                if crons:
                    self.crons[qname] = crons

    @defer.inlineCallbacks
    def stopService(self):
        for queue, crons in self.crons.items():
            runner = yield self.checkCronRunners(queue)

            if runner == self.uuid:
                log.msg('Deregistering queue %s' % queue)
                yield self.deregisterCronRunner(queue)

    @defer.inlineCallbacks
    def startBackend(self):
        try:
            self.client = getattr(importlib.import_module(
                self.backend), 'Backend')(self.config)
            yield self.client.connect()

        except exceptions.ImportError, e:
            raise Exception("Unable to load backend %s" % self.backend)

    @defer.inlineCallbacks
    def startService(self):
        yield self.startBackend()
        
        log.msg('Starting Rhumba')
    
        reactor.callWhenRunning(self.startBeat)
        self.setupQueues()
        queues = 0
        for k, v in self.queues.items():
            if v.plugin:
                queues += 1
                log.msg('Starting queue %s: plugin=%s' % (k, v.plugin))
                reactor.callWhenRunning(v.startQueue)

        if queues < 1:
            log.msg('No queues are running')

        if self.api_service:
            self.api = APIService(self.config, self)

            reactor.callWhenRunning(self.api.startAPI)

def makeService(config):
    return RhumbaService(config)




import json

from StringIO import StringIO

from zope.interface import implements

from twisted.internet import reactor, protocol, defer, error
from twisted.web.http_headers import Headers
from twisted.web.iweb import IBodyProducer
from twisted.web.client import Agent
from twisted.python import log


class ProcessProtocol(protocol.ProcessProtocol):
    """ProcessProtocol which supports timeouts"""
    def __init__(self, deferred, timeout):
        self.timeout = timeout
        self.timer = None

        self.deferred = deferred
        self.outBuf = StringIO()
        self.errBuf = StringIO()
        self.outReceived = self.outBuf.write
        self.errReceived = self.errBuf.write

    def processEnded(self, reason):
        if self.timer and (not self.timer.called):
            self.timer.cancel()

        out = self.outBuf.getvalue()
        err = self.errBuf.getvalue()

        e = reason.value
        code = e.exitCode

        if e.signal:
            self.deferred.errback(reason)
        else:
            self.deferred.callback((out, err, code))

    def connectionMade(self):
        @defer.inlineCallbacks
        def killIfAlive():
            try:
                yield self.transport.signalProcess('KILL')
                log.msg('Killed proccess: Timeout %s exceeded' % self.timeout)
            except error.ProcessExitedAlready:
                pass

        self.timer = reactor.callLater(self.timeout, killIfAlive)

def fork(executable, args=(), env={}, path=None, timeout=3600):
    """fork
    Provides a deferred wrapper function with a timeout function

    :param executable: Executable
    :type executable: str.
    :param args: Tupple of arguments
    :type args: tupple.
    :param env: Environment dictionary
    :type env: dict.
    :param timeout: Kill the child process if timeout is exceeded
    :type timeout: int.
    """
    d = defer.Deferred()
    p = ProcessProtocol(d, timeout)
    reactor.spawnProcess(p, executable, (executable,)+tuple(args), env, path)
    return d




import redis
import time
import json
import uuid

from twisted.internet import reactor, defer, protocol
from twisted.python import log

from txredis.client import RedisClient

from rhumba.backend import RhumbaBackend

class Backend(RhumbaBackend):
    """
    Rhumba redis backend
    """
    def __init__(self, config):
        self.config = config
        self.redis_host = self.config.get('redis_host', 'localhost')
        self.redis_port = int(self.config.get('redis_port', 6379))
        self.redis_db = int(self.config.get('redis_db', 0))

        self.client = None

    @defer.inlineCallbacks
    def connect(self):
        clientCreator = protocol.ClientCreator(
            reactor, RedisClient, db=self.redis_db)

        self.client = yield clientCreator.connectTCP(
            self.redis_host, self.redis_port
        )

    @defer.inlineCallbacks
    def queue(self, queue, message, params={}, uids=[]):
        """
        Queue a job in Rhumba
        """
        log.msg('Queing job %s/%s:%s' % (queue, message, repr(params)))

        d = {
            'id': uuid.uuid1().get_hex(),
            'version': 1,
            'message': message,
            'params': params
        }
        
        if uids:
            for uid in uids:
                yield self.client.lpush('rhumba.dq.%s.%s' % (
                    uid, queue), json.dumps(d))
        else:
            yield self.client.lpush('rhumba.q.%s' % queue, json.dumps(d))

        defer.returnValue(d['id'])

    @defer.inlineCallbacks
    def popQueue(self, queue):
        item = yield self.client.rpop("rhumba.q.%s" % queue)
        if item:
            defer.returnValue(json.loads(item))
        else:
            defer.returnValue(None)

    @defer.inlineCallbacks
    def popDirectQueue(self, name, queue):
        item = yield self.client.rpop("rhumba.dq.%s.%s" % (name, queue))
        if item:
            defer.returnValue(json.loads(item))
        else:
            defer.returnValue(None)

    @defer.inlineCallbacks
    def getResult(self, queue, uid, suid=None):

        if suid:
            r = yield self.client.get('rhumba.dq.%s.%s.%s' % (suid, queue, uid))
        else:
            r = yield self.client.get('rhumba.q.%s.%s' % (queue, uid))

        if r:
            defer.returnValue(json.loads(r))
        else:
            defer.returnValue([])

    def waitForResult(self, queue, uid, timeout=3600, suid=None):
        d = defer.Deferred()

        t = time.time()

        def checkResult():
            def result(r):
                if r:
                    return d.callback(r)

                if (time.time() - t) > timeout:
                    raise Exception(
                        "Timeout waiting for result on %s:%s" % (queue, uid))
                else:
                    reactor.callLater(1, checkResult)

            self.getResult(queue, uid, suid=suid
                ).addCallback(result)

        reactor.callLater(0, checkResult)

        return d

    def get(self, key):
        return self.client.get(key)

    def set(self, key, value, expire=None):
        return self.client.set(key, value, expire=expire)

    def keys(self, pattern):
        return self.client.keys(pattern)

    def queueSize(self, queue):
        return self.client.llen("rhumba.q.%s" % queue)
    
    def delete(self, key):
        return self.client.delete(key)

    @defer.inlineCallbacks
    def queueStats(self, queue, message, duration):
        yield self.client.incr(
            'rhumba.qstats.%s.%s.time' % (queue, message),
            int(duration*100000)
        )

        yield self.client.incr(
            'rhumba.qstats.%s.%s.count' % (queue, message))

    @defer.inlineCallbacks
    def getQueueMessageStats(self, queue):
        keys = yield self.client.keys('rhumba.qstats.%s.*' % queue)
        msgstats = {}
        for key in keys:
            msg = key.split('.')[3]
            stat = key.split('.')[4]
            val = yield self.get(key)

            print key, val
            if stat == 'time':
                val = int(val)/100.0
            else:
                val = int(val)

            if msg in msgstats:
                msgstats[msg][stat] = val
            else:
                msgstats[msg] = {stat: val}

        defer.returnValue(msgstats)

    @defer.inlineCallbacks
    def clusterQueues(self):
        """ Return a dict of queues in cluster and servers running them
        """
        servers = yield self.keys('rhumba\.server\.*\.heartbeat')

        queues = {}

        for s in servers:
            sname = s.split('.', 2)[-1].rsplit('.', 1)[0]

            qs = yield self.get('rhumba.server.%s.queues' % sname)
            uuid = yield self.get('rhumba.server.%s.uuid' % sname)
       
            qs = json.loads(qs)

            for q in qs:
                if q not in queues:
                    queues[q] = []

                queues[q].append({'host': sname, 'uuid': uuid})

        defer.returnValue(queues)

    @defer.inlineCallbacks
    def clusterStatus(self):
        """
        Returns a dict of cluster nodes and their status information
        """
        servers = yield self.keys('rhumba\.server\.*\.heartbeat')

        d = {
            'workers': {},
            'crons': {},
            'queues': {}
        }

        now = time.time()

        reverse_map = {}

        for s in servers:
            sname = s.split('.', 2)[-1].rsplit('.', 1)[0]

            last = yield self.get('rhumba.server.%s.heartbeat' % sname)
            status = yield self.get('rhumba.server.%s.status' % sname)
            uuid = yield self.get('rhumba.server.%s.uuid' % sname)

            reverse_map[uuid] = sname

            if not last:
                last = 0

            last = float(last)

            if (status == 'ready') and (now - last > 5):
                status = 'offline'

            if not sname in d['workers']:
                d['workers'][sname] = []

            d['workers'][sname].append({
                'lastseen': last,
                'status': status,
                'id': uuid
            })

        # Crons
        crons = yield self.keys('rhumba\.crons\.*')

        for key in crons:
            segments = key.split('.')

            queue = segments[2]
            if queue not in d['crons']:
                d['crons'][queue] = {'methods': {}}

            if len(segments)==4:
                last = yield self.get(key)
                d['crons'][queue]['methods'][segments[3]] = float(last)
            else:
                uid = yield self.get(key)
                d['crons'][queue]['master'] = '%s:%s' % (uid, reverse_map[uid])

        # Queues
        queue_keys = yield self.keys('rhumba.qstats.*')

        for key in queue_keys:
            qname = key.split('.')[2]
            if qname not in d['queues']:
                qlen = yield self.queueSize(qname)

                stats = yield self.getQueueMessageStats(qname)

                d['queues'][qname] = {
                    'waiting': qlen,
                    'messages': stats
                }

        defer.returnValue(d)




# -*- coding: utf-8 -*-
import json
import cgi
import hmac
import hashlib
import base64
import re
import time

from twisted.web import server, resource
from twisted.internet import defer, reactor
from twisted.python import log

try:
    from urllib.parse import urlparse
except:
    from urlparse import urlparse

try:
    from twisted.internet import ssl

    SSL=True
except:
    SSL=False


class APIProcessor(object):
    def __init__(self, service, config):
        self.service = service
        self.config = config

    def queue_detail(self, request, queue):
        return {
            'name': queue
        }

    @defer.inlineCallbacks
    def queue_call(self, request, queue, method):
        params = request.data

        id = yield self.service.client.queue(queue, method, params)
        defer.returnValue({'uid': id})

    def queue_wait_result(self, request, queue, uid):
        return self.service.client.waitForResult(queue, uid)

    @defer.inlineCallbacks
    def queue_call_wait(self, request, queue, method):
        call = yield self.queue_call(request, queue, method)
        result = yield self.queue_wait_result(request, queue, call['uid'])

        defer.returnValue(result)

    def list_queues(self, request):
        log.msg('List queues' + repr(request))

        return self.service.queues.keys()

    def cluster_detail(self, request):
        return self.service.client.clusterStatus()

class APIResource(resource.Resource):
    isLeaf = True
    addSlash = True

    def __init__(self, config, service):
        self.config = config
        self.service = service
        self.rhumba_service = service.service

        api = APIProcessor(self.rhumba_service, self.config)

        self.paths = (
            (r'^/queues/(\w+)/call/(\w+)', api.queue_call),
            (r'^/queues/(\w+)/result/(\w+)', api.queue_wait_result),
            (r'^/queues/(\w+)/wait/(\w+)', api.queue_call_wait),
            (r'^/queues/(\w+)', api.queue_detail),
            (r'^/queues/$', api.list_queues),
            (r'^/cluster/$', api.cluster_detail),
        )

    def completeCall(self, response, request):
        # Render the json response from call
        response = json.dumps(response)
        request.write(response)
        request.finish()

    def getProcessor(self, path):
        for regex, processor in self.paths:
            m = re.match(regex, path)
            if m:
                return m.groups(), processor
        
        return [], []

    def jsonRequest(self, path, request, data={}):
        params, proc = self.getProcessor(path)
        request.data = data
        if proc:
            return proc(request, *params)
        else:
            return None
        
    def getHeader(self, request, header, default=None):
        head = request.requestHeaders.getRawHeaders(header)
        if head:
            return head[0]
        else:
            return default

    def getSecret(self, auth):
        if auth == self.service.api_token:
            return self.service.api_secret

    def checkSignature(self, request, data=None):
        auth = self.getHeader(request, 'authorization', None)
        sig = self.getHeader(request, 'sig', None)

        if not (auth and sig):
            return False

        sign = [auth, request.method, request.path]
        if data:
            sign.append(
                hashlib.sha1(data).hexdigest()
            )

        key = self.getSecret(auth)

        if key:
            mysig = hmac.new(
                key=key,
                msg='\n'.join(sign),
                digestmod=hashlib.sha1
            ).digest()

            return base64.b64encode(mysig) == sig
        else:
            return False

    def render_GET(self, request):
        request.setHeader("content-type", "application/json")

        if self.service.api_token: 
            if not self.checkSignature(request):
                return '["Not authorized"]'

        d = defer.maybeDeferred(self.jsonRequest, request.path, request)

        d.addCallback(self.completeCall, request)

        return server.NOT_DONE_YET

    def render_POST(self, request):
        request.setHeader("content-type", "application/json")
        # Get request
        data = cgi.escape(request.content.read())

        if self.service.api_token: 
            if not self.checkSignature(request, data):
                return '["Not authorized"]'

        d = defer.maybeDeferred(self.jsonRequest, request.path, request,
            json.loads(data)
        )

        d.addCallback(self.completeCall, request)

        return server.NOT_DONE_YET

class APIService(object):
    def __init__(self, config, service):
        self.config = config
        self.service = service

        self.api_port = int(config.get('api_port', 7701))
        self.api_token = config.get('api_token', None)
        self.api_secret = config.get('api_secret', None)
        self.api_ssl = config.get('api_ssl', False)
        self.api_ssl_cert = config.get('api_ssl_cert', None)
        self.api_ssl_key = config.get('api_ssl_key', None)

    def startAPI(self):
        site = server.Site(APIResource(self.config, self))

        if self.api_ssl:
            if SSL:
                reactor.listenSSL(self.api_port, site,
                    ssl.DefaultOpenSSLContextFactory(
                        self.api_ssl_key, self.api_ssl_cert
                    )
                )
            else:
                raise Exception("Unable to start SSL API service, no OpenSSL")

        else:
            reactor.listenTCP(self.api_port, site)




import datetime
import calendar

Sunday = 0
Monday = 1
Tuesday = 2
Wednesday = 3
Thursday = 4
Friday = 5
Saturday = 6

Weekdays = (1, 2, 3, 4, 5)

class Cronable(object):
    def __init__(self, name, secs, min, hour, day, month, weekday):
        self.name = name
        self.secs = secs
        self.min = min
        self.hour = hour
        self.day = day
        self.month = month
        self.weekday = weekday

        self.chk = [
            (self.secs, self._checkSecs),
            (self.min, self._checkMins),
            (self.hour, self._checkHours),
            (self.day, self._checkDays),
            (self.weekday, self._checkWeekday),
            (self.month, self._checkMonth),
        ]

        self.ignore_delta = False
        for a, b in self.chk:
            if isinstance(a, str) and (a.startswith('*/')):
                self.ignore_delta = True

    def _checkEvery(self, t, n, delta, cur, wrap=59, exact=True):
        if not t:
            return False
        tseg = t

        if isinstance(t, str):
            if t.startswith('*/'):
                tseg = int(t[2:])
                if delta >= tseg:
                    return True
                return False
            elif '-' in t:
                ra_start, ra_stop = [int(i) for i in t.split('-')]
                tseg = range(ra_start, ra_stop+1)
            else:
                tseg = int(t)

        if (isinstance(tseg, list) or isinstance(tseg, tuple)):
            if (n in tseg) and ((delta >= 1) or self.ignore_delta or cur):
                return True

        delmatch = (delta >= wrap) or self.ignore_delta or cur

        if exact:
            if (tseg == n) and delmatch:
                return True

        else:
            if (n >= tseg) and delmatch:
                return True

        return False

    def _checkSecs(self, now, delta, cur):
        delta_secs = delta.total_seconds()

        return self._checkEvery(
            self.secs, now.second, delta_secs, cur, exact=False)

    def _checkMins(self, now, delta, cur):
        delta_mins = delta.total_seconds() / 60

        return self._checkEvery(
            self.min, now.minute, delta_mins, cur, exact=False)

    def _checkHours(self, now, delta, cur):
        delta_hours = delta.total_seconds() / (60*60)

        return self._checkEvery(self.hour, now.hour, delta_hours, cur, wrap=23)
    
    def _checkDays(self, now, delta, cur):
        delta_days = delta.total_seconds() / (60*60*24)

        # Get the max days in this month
        start, end = calendar.monthrange(now.year, now.month)

        return self._checkEvery(self.day, now.day, delta_days, cur, wrap=end)

    def _checkMonth(self, now, delta, cur):
        # Get the max days in this month
        if now.month == 1:
            lastm = 12
            year = now.year - 1
        else:
            lastm = now.month - 1
            year = now.year

        start, end = calendar.monthrange(year, lastm)

        delta_days = delta.total_seconds() / (60*60*24*end)

        return self._checkEvery(self.month, now.month, delta_days, cur,
            wrap=12, exact=True)

    def _checkWeekday(self, now, delta, cur):
        if isinstance(self.weekday, str):
            self.weekday = int(self.weekday)

        today = now.isoweekday()

        if isinstance(self.weekday, tuple) or isinstance(self.weekday, list):
            if (today in self.weekday):
                if (delta.days >= 1) or self.ignore_delta or cur:
                    return True

        elif self.weekday == today:
            if ((delta.days > 7) or self.ignore_delta or cur):
                return True

        return False

    def checkCron(self, last, now):
        assert(isinstance(now, datetime.datetime))

        if not last:
            last = 0 

        last_date = datetime.datetime.fromtimestamp(float(last))
        delta = now - last_date

        every = False
        run = True
        r = False

        matches = []
        for a, b in self.chk:
            if a:
                r = b(now, delta, r)
                matches.append(r)
                run = run and r

        return run
            

def cron(secs=None, min=None, hour=None, day=None, month=None, weekday=None):
    def wrapper(fn):
        fn.cronable = Cronable(
            fn.__name__, secs, min, hour, day, month, weekday)
        return fn
    return wrapper



import json

from StringIO import StringIO

from zope.interface import implements

from twisted.internet import reactor, protocol, defer, error
from twisted.web.http_headers import Headers
from twisted.web.iweb import IBodyProducer
from twisted.web.client import Agent
from twisted.python import log

try:
    from twisted.internet.ssl import ClientContextFactory

    class WebClientContextFactory(ClientContextFactory):
        def getContext(self, hostname, port):
            return ClientContextFactory.getContext(self)
    SSL=True
except:
    SSL=False

try:
    from twisted.web import client
    client._HTTP11ClientFactory.noisy = False
    client.HTTPClientFactory.noisy = False
except:
    pass


class Timeout(Exception):
    """
    Raised to notify that an operation exceeded its timeout.
    """

class BodyReceiver(protocol.Protocol):
    """ Simple buffering consumer for body objects """
    def __init__(self, finished):
        self.finished = finished
        self.buffer = StringIO()

    def dataReceived(self, buffer):
        self.buffer.write(buffer)

    def connectionLost(self, reason):
        self.buffer.seek(0)
        self.finished.callback(self.buffer)

class StringProducer(object):
    """String producer for writing to HTTP requests
    """
    implements(IBodyProducer)

    def __init__(self, body):
        self.body = body
        self.length = len(body)

    def startProducing(self, consumer):
        consumer.write(self.body)
        return defer.succeed(None)

    def pauseProducing(self):
        pass

    def stopProducing(self):
        pass

class HTTPRequest(object):
    def abort_request(self, request):
        """Called to abort request on timeout"""
        self.timedout = True
        try:
            request.cancel()
        except error.AlreadyCancelled:
            return

    @defer.inlineCallbacks
    def response(self, request):
        if request.length:
            d = defer.Deferred()
            request.deliverBody(BodyReceiver(d))
            b = yield d
            body = b.read()
        else:
            body = ""

        defer.returnValue(body)

    def request(self, url, method='GET', headers={}, data=None, socket=None, timeout=120):
        if socket:
            agent = SocketyAgent(reactor, socket)
        else:
            if url[:5] == 'https':
                if SSL:
                    agent = Agent(reactor, WebClientContextFactory())
                else:
                    raise Exception('HTTPS requested but not supported')
            else:
                agent = Agent(reactor)

        request = agent.request(method, url,
            Headers(headers),
            StringProducer(data) if data else None
        )

        if timeout:
            timer = reactor.callLater(timeout, self.abort_request,
                request)

            def timeoutProxy(request):
                if timer.active():
                    timer.cancel()
                return self.response(request)
            def requestAborted(failure):
                if timer.active():
                    timer.cancel()
    
                failure.trap(defer.CancelledError,
                             error.ConnectingCancelledError)
    
                raise Timeout(
                    "Request took longer than %s seconds" % timeout)

            request.addCallback(timeoutProxy).addErrback(requestAborted)
        else:
            request.addCallback(self.response)

        return request

    @classmethod
    def getBody(cls, url, method='GET', headers={}, data=None, socket=None, timeout=120):
        """Make an HTTP request and return the body
        """
        if not 'User-Agent' in headers:
            headers['User-Agent'] = ['Tensor HTTP checker']

        return cls().request(url, method, headers, data, socket, timeout)

    @classmethod
    @defer.inlineCallbacks
    def getJson(cls, url, method='GET', headers={}, data=None, socket=None, timeout=120):
        """Fetch a JSON result via HTTP
        """
        if not 'Content-Type' in headers:
            headers['Content-Type'] = ['application/json']

        body = yield cls().getBody(url, method, headers, data, socket, timeout)

        defer.returnValue(json.loads(body))




import json
import time

from twisted.trial import unittest

from twisted.internet import defer, reactor
from twisted.web import server, resource
from twisted.python import log

from rhumba import http_api
from rhumba.http_client import HTTPRequest

class TestResource(resource.Resource):
    isLeaf = True
    addSlash = True

    def completeCall(self, request):
        request.write("")
        try:
            request.finish()
        except:
            pass

    def render_GET(self, request):
        # Takes 5 seconds to complete
        reactor.callLater(2.5, self.completeCall, request)

        return server.NOT_DONE_YET

class RhumbaTest(unittest.TestCase):
    def sleep(self, secs):
        d = defer.Deferred()
        reactor.callLater(secs, d.callback, None)
        return d

    @defer.inlineCallbacks
    def setUp(self):
        site = server.Site(TestResource())
        self.con = yield reactor.listenTCP(1234, site)

    @defer.inlineCallbacks
    def tearDown(self):
        yield self.con.loseConnection()
        yield self.sleep(2.5)

    @defer.inlineCallbacks
    def test_client_timeout(self):
        t = time.time()
        try:
            d = yield HTTPRequest.getBody('http://localhost:1234/', timeout=1)
        except:
            pass

        l = time.time() - t

        self.assertLess(l, 2)



from rhumba.http_client import HTTPRequest

from twisted.python import log


class RhumbaPlugin(object):
    """
    Simple object for Rhumba plugins
    """
    def __init__(self, config, client):
        self.config = config
        self.client = client
        self.queue_name = config['name']
        self.http = HTTPRequest

    def log(self, *a, **kw):
        log.msg("[%s]: %s" % (self.config['name'], a[0]), *a[1:], **kw)




from twisted.internet import defer, reactor
from rhumba import RhumbaPlugin, cron

def sleep(secs):
    d = defer.Deferred()
    reactor.callLater(secs, d.callback, None)
    return d

class Plugin(RhumbaPlugin):
    @defer.inlineCallbacks
    def call_test(self, args):
        self.log("Test call %s" % repr(args))
        yield sleep(2)

        name = args.get('name')
        if name:
            defer.returnValue(["Hello, %s!" % name])
        else:
            defer.returnValue(["Hello!"])

    @cron(secs="*/5")
    @defer.inlineCallbacks
    def call_crontest(self, args):
        qs = yield self.client.clusterQueues()

        self.log(repr(qs[self.queue_name]))
        self.log("tick!")




import time
import exceptions
import json
import importlib

from twisted.internet import task, reactor, defer
from twisted.python import log


class RhumbaQueue(object):
    def __init__(self, config, svc):
        self.config = config
        self.service = svc

        self.name = config['name']
        self.expire = int(config.get('expire', 3600))
        self.plugin = self.loadPlugin(config['plugin'])

        self.inter = float(config.get('inter', 1))

        self.fast_inter = float(config.get('fast_inter', 0.1))

        self.max_jobs = config.get('max_jobs', 5)
        self.jobs = 0
        self.cycl = 0

        self.t = task.LoopingCall(self.tick)

    def loadPlugin(self, plugin):
        try:
            return getattr(importlib.import_module(plugin), 'Plugin')(
                self.config, self.service.client)

        except exceptions.ImportError, e:
            log.msg("Error importing plugin %s : %s" % (plugin, repr(e)))
            return None
        
    @defer.inlineCallbacks
    def processQueue(self, item, result_key='q'):
        m = item.get('message')
        if m:
            uid = item['id']
            yield self.service.setStatus("processing:%s:%s:%s" % (
                m, uid, time.time()))

            start = time.time()
            result = yield self.processItem(m, item.get('params', {}))
            duration = time.time() - start

            d = {
                'result': result,
                'time': time.time()
            }

            yield self.service.client.queueStats(self.name, m, duration)

            yield self.service.client.set(
                'rhumba.%s.%s.%s' % (result_key, self.name, uid),
                json.dumps(d), expire=self.expire)

    @defer.inlineCallbacks
    def reQueue(self, request):
        response = yield self.service.client.lpush("rhumba.q.%s" % self.name)

    def processItem(self, message, params):
        fn = getattr(self.plugin, 'call_%s' % message)

        return defer.maybeDeferred(fn, params)

    @defer.inlineCallbacks
    def queueRun(self):
        if self.cycl > 0:
            self.cycl -= 1
            defer.returnValue(None)

        if self.jobs >= self.max_jobs:
            defer.returnValue(None)

        item = yield self.service.client.popQueue(self.name)

        if item:
            self.jobs += 1
            yield self.processQueue(item)
            self.jobs -= 1
        else:
            self.cycl = self.inter/self.fast_inter

            yield self.service.setStatus("ready")

        defer.returnValue(None)

    @defer.inlineCallbacks
    def queueFan(self):

        item = yield self.service.client.popDirectQueue(self.service.uuid, self.name)

        if item:
            yield self.processQueue(item, result_key='dq.%s' % self.service.uuid)
        else:
            yield self.service.setStatus("ready")

        defer.returnValue(None)

    def tick(self):
        reactor.callLater(self.fast_inter, self.queueRun)
        reactor.callLater(self.fast_inter, self.queueFan)

    def startQueue(self):
        """Starts the timer for this queue"""
        self.td = self.t.start(self.fast_inter)

    def stopQueue(self):
        """Stops the timer for this queue"""
        self.td = None
        self.t.stop()
